@@ -93,6 +93,27 @@ class PackedMethodBitField:
     def full(self):
         return self.count() == 8
 
+def multiLineFormat(things, prologue, separator, lineSeparator, epilogue, thingsPerLine = 4):
+    r = [prologue]
+    i = 0
+    for t in things:
+        if i != 0:
+            if i % thingsPerLine == 0:
+                r += [lineSeparator]
+            else:
+                r += [separator]
+        r += [t]
+        i += 1
+    r += [epilogue]
+    return "".join(r)
+
+def prettyType(typeName, subTypes, typesPerLine = 4):
+    """Pretty print a type signature made up of many alternative subtypes"""
+    sTs = multiLineFormat(subTypes,
+                          "( ", " | ", "\n       | ", " )",
+                          thingsPerLine = typesPerLine)
+    return "-type(%s ::\n       %s)." % (typeName, sTs)
+
 def printFileHeader():
     print """%%   Autogenerated code. Do not edit.
 %%
@@ -126,7 +147,7 @@ def printFileHeader():
 %%
 %%   Contributor(s): ______________________________________.
 %%"""
-    
+
 def genErl(spec):
     def erlType(domain):
         return erlangTypeMap[spec.resolveDomain(domain)]
@@ -151,7 +172,7 @@ def genErl(spec):
 
     def genMethodHasContent(m):
         print "method_has_content(%s) -> %s;" % (m.erlangName(), str(m.hasContent).lower())
-    
+
     def genMethodIsSynchronous(m):
         hasNoWait = "nowait" in fieldNameList(m.arguments)
         if m.isSynchronous and hasNoWait:
@@ -219,6 +240,9 @@ def genErl(spec):
             else:
                 pass
 
+    def genMethodRecord(m):
+        print "method_record(%s) -> #%s{};" % (m.erlangName(), m.erlangName())
+
     def genDecodeMethodFields(m):
         packedFields = packMethodFields(m.arguments)
         binaryPattern = ', '.join([methodFieldFragment(f) for f in packedFields])
@@ -299,6 +323,7 @@ def genErl(spec):
 -export([method_id/1]).
 -export([method_has_content/1]).
 -export([is_method_synchronous/1]).
+-export([method_record/1]).
 -export([method_fieldnames/1]).
 -export([decode_method_fields/2]).
 -export([decode_properties/2]).
@@ -310,6 +335,22 @@ def genErl(spec):
 bitvalue(true) -> 1;
 bitvalue(false) -> 0;
 bitvalue(undefined) -> 0.
+
+%% Method signatures
+-ifdef(use_specs).
+-spec(lookup_method_name/1 :: (amqp_method()) -> amqp_method_name()).
+-spec(method_id/1 :: (amqp_method_name()) -> amqp_method()).
+-spec(method_has_content/1 :: (amqp_method_name()) -> boolean()).
+-spec(is_method_synchronous/1 :: (amqp_method_record()) -> boolean()).
+-spec(method_record/1 :: (amqp_method_name()) -> amqp_method_record()).
+-spec(method_fieldnames/1 :: (amqp_method_name()) -> [amqp_method_field_name()]).
+-spec(decode_method_fields/2 :: (amqp_method_name(), binary()) -> amqp_method_record()).
+-spec(decode_properties/2 :: (non_neg_integer(), binary()) -> amqp_property_record()).
+-spec(encode_method_fields/1 :: (amqp_method_record()) -> binary()).
+-spec(encode_properties/1 :: (amqp_method_record()) -> binary()).
+-spec(lookup_amqp_exception/1 :: (amqp_exception()) -> {boolean(), amqp_exception_code(), binary()}).
+-spec(amqp_exception/1 :: (amqp_exception_code()) -> amqp_exception()).
+-endif. % use_specs
 """
     for m in methods: genLookupMethodName(m)
     print "lookup_method_name({_ClassId, _MethodId} = Id) -> exit({unknown_method_id, Id})."
@@ -323,6 +364,9 @@ bitvalue(undefined) -> 0.
     for m in methods: genMethodIsSynchronous(m)
     print "is_method_synchronous(Name) -> exit({unknown_method_name, Name})."
 
+    for m in methods: genMethodRecord(m)
+    print "method_record(Name) -> exit({unknown_method_name, Name})."
+
     for m in methods: genMethodFieldNames(m)
     print "method_fieldnames(Name) -> exit({unknown_method_name, Name})."
 
@@ -362,7 +406,7 @@ def genHrl(spec):
                 result += ' = ' + conv_fn(field.defaultvalue)
             return result
         return ', '.join([fillField(f) for f in fields])
-    
+
     methods = spec.allMethods()
 
     printFileHeader()
@@ -381,12 +425,75 @@ def genHrl(spec):
     for c in spec.allClasses():
         print "-record('P_%s', {%s})." % (erlangize(c.name), fieldNameList(c.fields))
 
+    print "-ifdef(use_specs)."
+    print "%% Various types"
+    print prettyType("amqp_method_name()",
+                     [m.erlangName() for m in methods])
+    print prettyType("amqp_method()",
+                     ["{%s, %s}" % (m.klass.index, m.index) for m in methods],
+                     6)
+    print prettyType("amqp_method_record()",
+                     ["#%s{}" % (m.erlangName()) for m in methods])
+    fieldNames = set()
+    for m in methods:
+        fieldNames.update(m.arguments)
+    fieldNames = [erlangize(f.name) for f in fieldNames]
+    print prettyType("amqp_method_field_name()",
+                     fieldNames)
+    print prettyType("amqp_property_record()",
+                     ["#'P_%s'{}" % erlangize(c.name) for c in spec.allClasses()])
+    print prettyType("amqp_exception()",
+                     ["'%s'" % erlangConstantName(c).lower() for (c, v, cls) in spec.constants])
+    print prettyType("amqp_exception_code()",
+                     ["%i" % v for (c, v, cls) in spec.constants])
+    print "-endif. % use_specs"
+
+def genSpec(spec):
+    methods = spec.allMethods()
+
+    printFileHeader()
+    print """% Hard-coded types
+-type(amqp_field_type() ::
+      'longstr' | 'signedint' | 'decimal' | 'timestamp' |
+      'table' | 'byte' | 'double' | 'float' | 'long' |
+      'short' | 'bool' | 'binary' | 'void').
+-type(amqp_property_type() ::
+      'shortstr' | 'longstr' | 'octet' | 'shortint' | 'longint' |
+      'longlongint' | 'timestamp' | 'bit' | 'table').
+%% we could make this more precise but ultimately are limited by
+%% dialyzer's lack of support for recursive types
+-type(amqp_table() :: [{binary(), amqp_field_type(), any()}]).
+%% TODO: make this more precise
+-type(amqp_properties() :: tuple()).
+
+-type(channel_number() :: non_neg_integer()).
+-type(resource_name() :: binary()).
+-type(routing_key() :: binary()).
+-type(username() :: binary()).
+-type(password() :: binary()).
+-type(vhost() :: binary()).
+-type(ctag() :: binary()).
+-type(exchange_type() :: atom()).
+-type(binding_key() :: binary()).
+"""
+    print "% Auto-generated types"
+    classIds = set()
+    for m in spec.allMethods():
+        classIds.add(m.klass.index)
+    print prettyType("amqp_class_id()",
+                     ["%i" % ci for ci in classIds])
+
 def generateErl(specPath):
     genErl(AmqpSpec(specPath))
 
 def generateHrl(specPath):
     genHrl(AmqpSpec(specPath))
-    
+
+def generateSpec(specPath):
+    genSpec(AmqpSpec(specPath))
+
 if __name__ == "__main__":
-    do_main(generateHrl, generateErl)
+    do_main_dict({"header": generateHrl,
+                  "spec": generateSpec,
+                  "body": generateErl})
 
