@@ -443,7 +443,7 @@ init_it(Starter, Parent, Name0, Mod, Args, Options) ->
 name({local,Name}) -> Name;
 name({global,Name}) -> Name;
 %% name(Pid) when is_pid(Pid) -> Pid;
-%% when R11 goes away, drop the line beneath and uncomment the line above
+%% when R12 goes away, drop the line beneath and uncomment the line above
 name(Name) -> Name.
 
 unregister_name({local,Name}) ->
@@ -607,9 +607,9 @@ process_msg(Parent, Name, State, Mod, Time, TimeoutState, Queue,
             Debug, Msg) ->
     case Msg of
 	{system, From, Req} ->
-	    sys:handle_system_msg
-              (Req, From, Parent, ?MODULE, Debug,
-               [Name, State, Mod, Time, TimeoutState, Queue]);
+	    sys:handle_system_msg(
+              Req, From, Parent, ?MODULE, Debug,
+              [Name, State, Mod, Time, TimeoutState, Queue]);
         %% gen_server puts Hib on the end as the 7th arg, but that
         %% version of the function seems not to be documented so
         %% leaving out for now.
@@ -639,7 +639,7 @@ do_multi_call(Nodes, Name, Req, Timeout) ->
     Caller = self(),
     Receiver =
 	spawn(
-	  fun() ->
+	  fun () ->
 		  %% Middleman process. Should be unsensitive to regular
 		  %% exit signals. The sychronization is needed in case
 		  %% the receiver would exit before the caller started
